import typing as t
from unittest.mock import AsyncMock, MagicMock

import pytest
from anyio import Path as AsyncPath
from jinja2.environment import Template
from jinja2.exceptions import TemplateNotFound
from jinja2_async_environment.environment import AsyncEnvironment
from jinja2_async_environment.loaders import (
    AsyncChoiceLoader,
    AsyncDictLoader,
    AsyncFileSystemLoader,
)


class TestAsyncChoiceLoader:
    @pytest.fixture
    def template_dict1(self) -> dict[str, str]:
        return {
            "index.html": "<h1>Hello from dict1</h1>",
            "shared.html": "<p>Shared template from dict1</p>",
        }

    @pytest.fixture
    def template_dict2(self) -> dict[str, str]:
        return {
            "about.html": "<h1>About page from dict2</h1>",
            "shared.html": "<p>Shared template from dict2</p>",
        }

    @pytest.fixture
    def dict_loader1(self, template_dict1: dict[str, str]) -> AsyncDictLoader:
        return AsyncDictLoader(template_dict1, AsyncPath("/templates/dict1"))

    @pytest.fixture
    def dict_loader2(self, template_dict2: dict[str, str]) -> AsyncDictLoader:
        return AsyncDictLoader(template_dict2, AsyncPath("/templates/dict2"))

    @pytest.fixture
    def mock_fs_loader(self) -> AsyncFileSystemLoader:
        loader = MagicMock(spec=AsyncFileSystemLoader)
        loader.get_source_async = AsyncMock(side_effect=TemplateNotFound("not found"))
        loader.list_templates_async = AsyncMock(return_value=["fs_template.html"])
        return loader

    @pytest.fixture
    def choice_loader(
        self,
        dict_loader1: AsyncDictLoader,
        dict_loader2: AsyncDictLoader,
        mock_fs_loader: AsyncFileSystemLoader,
    ) -> AsyncChoiceLoader:
        return AsyncChoiceLoader(
            [dict_loader1, dict_loader2, mock_fs_loader], AsyncPath("/templates")
        )

    @pytest.fixture
    def environment(self, choice_loader: AsyncChoiceLoader) -> AsyncEnvironment:
        env = AsyncEnvironment(loader=choice_loader)
        env.enable_async = True
        return env

    def test_init(
        self,
        dict_loader1: AsyncDictLoader,
        dict_loader2: AsyncDictLoader,
        mock_fs_loader: AsyncFileSystemLoader,
    ) -> None:
        loaders = [dict_loader1, dict_loader2, mock_fs_loader]
        loader = AsyncChoiceLoader(loaders, AsyncPath("/templates"))
        assert loader.loaders == loaders
        assert len(loader.searchpath) == 1
        assert loader.searchpath[0] == AsyncPath("/templates")

    @pytest.mark.asyncio
    async def test_get_source_async_first_loader(
        self, choice_loader: AsyncChoiceLoader
    ) -> None:
        source, _, uptodate = await choice_loader.get_source_async("index.html")
        assert source == "<h1>Hello from dict1</h1>"
        assert callable(uptodate)

    @pytest.mark.asyncio
    async def test_get_source_async_second_loader(
        self, choice_loader: AsyncChoiceLoader
    ) -> None:
        source, _, uptodate = await choice_loader.get_source_async("about.html")
        assert source == "<h1>About page from dict2</h1>"
        assert callable(uptodate)

    @pytest.mark.asyncio
    async def test_get_source_async_shared_template(
        self, choice_loader: AsyncChoiceLoader
    ) -> None:
        source, _, uptodate = await choice_loader.get_source_async("shared.html")
        assert source == "<p>Shared template from dict1</p>"
        assert callable(uptodate)

    @pytest.mark.asyncio
    async def test_get_source_async_nonexistent_template(
        self, choice_loader: AsyncChoiceLoader
    ) -> None:
        with pytest.raises(TemplateNotFound):
            await choice_loader.get_source_async("nonexistent.html")

    @pytest.mark.asyncio
    async def test_get_source_async_with_path_object(
        self, choice_loader: AsyncChoiceLoader
    ) -> None:
        source, _, uptodate = await choice_loader.get_source_async(
            AsyncPath("index.html")
        )
        assert source == "<h1>Hello from dict1</h1>"
        assert callable(uptodate)

    @pytest.mark.asyncio
    async def test_list_templates_async(
        self, choice_loader: AsyncChoiceLoader, mock_fs_loader: AsyncFileSystemLoader
    ) -> None:
        templates = await choice_loader.list_templates_async()
        assert isinstance(templates, list)

        expected_templates = {
            "index.html",
            "about.html",
            "shared.html",
            "fs_template.html",
        }
        assert set(templates) == expected_templates
        assert templates == sorted(templates)

    @pytest.mark.asyncio
    async def test_load_async(
        self, choice_loader: AsyncChoiceLoader, environment: AsyncEnvironment
    ) -> None:
        template = await choice_loader.load_async(environment, "index.html")
        assert isinstance(template, Template)

        globals_dict: dict[str, t.Any] = {"name": "World"}
        template_with_globals = await choice_loader.load_async(
            environment, "index.html", globals_dict
        )
        assert isinstance(template_with_globals, Template)
        assert template_with_globals.globals.get("name") == "World"



import typing as t

from jinja2 import nodes
from jinja2.compiler import CodeGenerator, CompilerExit, Frame


class AsyncFrame(Frame):
    block_frame: "AsyncFrame | None"
    block_counters: t.Mapping[str, int]
    block_frame_id: int
    require_output_check: bool
    has_known_extends: bool
    toplevel: bool
    rootlevel: bool
    buffer: t.Any | None

    def copy(self) -> t.Self:
        rv = super().copy()
        t.cast(AsyncFrame, rv)
        return rv

    def inspect(self, nodes: t.Any | None = None) -> None: ...

    def push_scope(self) -> None: ...

    def pop_scope(self) -> None: ...

    def find_break(self) -> bool: ...

    def find_continue(self) -> bool: ...


class AsyncCodeGenerator(CodeGenerator):
    environment: t.Any
    name: str
    filename: str
    stream: t.Any
    extends_so_far: int
    has_known_extends: bool
    root_frame_class: t.Type[AsyncFrame] = AsyncFrame
    eval_ctx: t.Any = None

    def choose_async(self) -> str:  # type: ignore[override]
        return "async " if self.environment.enable_async else ""

    def simple_write(self, value: str, frame: Frame) -> None:  # type: ignore[override]
        self.writeline(f"yield {value}")

    def pull_locals(self, nodes: t.Any) -> None:
        pass

    def pull_dependencies(self, nodes: t.Any) -> None:
        pass

    def visit_Name(self, node: nodes.Name, frame: Frame) -> None:
        frame = t.cast(AsyncFrame, frame)
        if node.ctx == "store" and (
            frame.toplevel or frame.loop_frame or frame.block_frame
        ):
            if hasattr(self, "_assign_stack") and self._assign_stack:
                self._assign_stack[-1].add(node.name)

        if node.name in ("blocks", "debug_info"):
            self.write(node.name)
            return

        if node.ctx == "load":
            self.write(f"context.get({node.name!r})")
            return

        try:
            ref = frame.symbols.ref(node.name)

            if node.ctx == "load":
                from jinja2.compiler import VAR_LOAD_PARAMETER

                load = frame.symbols.find_load(ref)
                if not (
                    load is not None
                    and load[0] == VAR_LOAD_PARAMETER
                    and hasattr(self, "parameter_is_undeclared")
                    and not self.parameter_is_undeclared(ref)
                ):
                    self.write(
                        f"(undefined(name={node.name!r}) if {ref} is missing else {ref})"
                    )
                    return

            self.write(ref)
        except AssertionError:
            if node.ctx == "load":
                self.write(f"context.get({node.name!r}, undefined(name={node.name!r}))")
            else:
                self.write(f"context.vars[{node.name!r}]")

    def generate(self, node: nodes.Template) -> str:
        self.writeline(f"name = {self.name!r}")
        self.writeline("blocks = {}")
        self.writeline("debug_info = None")

        self.writeline("from jinja2.runtime import Undefined")
        self.writeline("def undefined(name=None, **_):")
        self.indent()
        self.writeline("return Undefined(name=name)")
        self.outdent()

        self.writeline("async def root(context):")
        self.indent()
        self.writeline("parent_template = None")
        self.writeline("environment = context.environment")
        self.writeline("eval_ctx = context.eval_ctx")
        self.writeline("missing = environment.undefined")

        from jinja2.nodes import EvalContext

        if self.eval_ctx is None:
            self.eval_ctx = EvalContext(self.environment, self.name)

        frame = self.root_frame_class(eval_ctx=self.eval_ctx)
        frame.toplevel = frame.rootlevel = True
        frame.require_output_check = False

        self.pull_locals(node.find_all(nodes.Scope))
        self.pull_dependencies(node.find_all((nodes.Import, nodes.FromImport)))
        self.blockvisit(node.body, frame)
        self.outdent()

        return self.stream.getvalue()

    def visit_For(self, node: nodes.For, frame: Frame) -> None:
        frame = t.cast(AsyncFrame, frame)
        node.iter_child_nodes(exclude=("iter",))
        if node.recursive:
            raise NotImplementedError("Recursive loops not supported")

        target = t.cast(t.Any, node.target)
        item = target.name

        frame.symbols.store(item)

        self.writeline(f"{item} = None")

        loop_filter = None
        if node.test:
            loop_filter = self.temporary_identifier()
            self.writeline(f"{loop_filter} = ", node.test)
            self.visit(node.test, frame)

        loop_var = self.temporary_identifier()
        self.writeline(f"{loop_var} = -1", node)

        self.writeline("try:", node)
        self.indent()
        self.writeline(f"async for {item} in ", node.iter)
        self.visit(node.iter, frame)
        self.write(":")
        self.indent()
        self.writeline(f"{loop_var} += 1")
        if node.test and loop_filter:
            self.writeline(f"if {loop_filter}({item}):")
            self.indent()
        self.blockvisit(node.body, frame)
        if node.test and loop_filter:
            self.outdent()
        self.outdent()
        self.outdent()
        self.writeline("except TypeError:", node)
        self.indent()

        self.writeline(f"for {item} in ", node.iter)
        self.visit(node.iter, frame)
        self.write(":")
        self.indent()
        self.writeline(f"{loop_var} += 1")
        if node.test and loop_filter:
            self.writeline(f"if {loop_filter}({item}):")
            self.indent()
        self.blockvisit(node.body, frame)
        if node.test and loop_filter:
            self.outdent()
        self.outdent()
        self.outdent()

        if node.else_:
            self.writeline(f"if {loop_var} == -1:")
            self.indent()
            self.blockvisit(node.else_, frame)
            self.outdent()

    def visit_Block(self, node: nodes.Block, frame: Frame) -> None:
        frame = t.cast(AsyncFrame, frame)

        block_name = node.name
        self.writeline(f"blocks[{block_name!r}] = []")

        block_func_name = f"block_{block_name}"
        self.writeline(f"{self.choose_async()}def {block_func_name}(context):")
        self.indent()

        self.writeline("yield ''")

        if node.body:
            self.blockvisit(node.body, frame)

        self.outdent()

        self.writeline(f"blocks[{block_name!r}].append({block_func_name})")

        level = 0
        if frame.toplevel:
            if self.has_known_extends:
                return
            if self.extends_so_far > 0:
                self.writeline("if parent_template is None:")
                self.indent()
                level += 1

        if node.scoped:
            context = self.derive_context(frame)
        else:
            context = self.get_context_ref()

        if node.required:
            self.writeline(f"if len(context.blocks[{node.name!r}]) <= 1:", node)
            self.indent()
            self.writeline(
                f'raise TemplateRuntimeError("Required block {node.name!r} not found")',
                node,
            )
            self.outdent()

        self.writeline("try:", node)
        self.indent()
        self.writeline(
            f"async for event in context.blocks[{node.name!r}][0]({context}):",
            node,
        )
        self.indent()
        self.simple_write("event", frame)
        self.outdent()
        self.outdent()
        self.writeline("except KeyError:", node)
        self.indent()
        self.writeline("yield ''")
        self.outdent()

        self.outdent(level)

    def visit_Extends(self, node: nodes.Extends, frame: Frame) -> None:
        frame = t.cast(AsyncFrame, frame)
        if not frame.require_output_check:
            raise CompilerExit()
        if not frame.toplevel:
            self.fail("cannot use extend from a non top-level scope", node.lineno)
        if self.extends_so_far > 0:
            if not self.has_known_extends:
                self.writeline("if parent_template is not None:")
                self.indent()
            self.writeline('raise TemplateRuntimeError("extended multiple times")')
            if self.has_known_extends:
                raise CompilerExit()
            else:
                self.outdent()
        self.writeline("parent_template = await environment.get_template_async(", node)
        self.visit(node.template, frame)
        self.write(f", {self.name!r})")
        self.writeline("for name, parent_block in parent_template.blocks.items():")
        self.indent()
        self.writeline("context.blocks.setdefault(name, []).append(parent_block)")
        self.outdent()
        if frame.rootlevel:
            self.has_known_extends = True
        self.extends_so_far += 1

    def visit_Include(self, node: nodes.Include, frame: Frame) -> None:
        frame = t.cast(AsyncFrame, frame)
        if node.ignore_missing:
            self.writeline("try:")
            self.indent()
        self.writeline("template = await environment.get_template_async(", node)
        self.visit(node.template, frame)
        self.write(f", {self.name!r})")
        if node.ignore_missing:
            self.outdent()
            self.writeline("except TemplateNotFound:")
            self.indent()
            self.writeline("pass")
            self.outdent()
            self.writeline("else:")
            self.indent()
        if node.with_context:
            self.writeline(
                f"async for event in template.root_render_func(template.new_context(context.get_all(), True, {self.dump_local_context(frame)})):"
            )
        else:
            self.writeline(
                "async for event in (await template._get_default_module_async())._body_stream:"
            )
        self.indent()
        self.simple_write("event", frame)
        self.outdent()
        if node.ignore_missing:
            self.outdent()

    def _import_common(
        self, node: nodes.Import | nodes.FromImport, frame: Frame
    ) -> None:
        frame = t.cast(AsyncFrame, frame)
        self.writeline("try:")
        self.indent()
        self.writeline("template = await environment.get_template_async(", node)
        self.visit(node.template, frame)
        self.write(f", {self.name!r})")
        self.outdent()
        self.writeline("except TemplateNotFound:")
        self.indent()
        self.writeline("template = await environment.get_template_async(", node)
        self.visit(node.template, frame)
        self.write(f", {self.name!r})")
        self.outdent()



import typing as t
from unittest.mock import MagicMock, patch

import pytest
from anyio import Path as AsyncPath
from jinja2.exceptions import TemplateNotFound
from jinja2_async_environment.environment import AsyncEnvironment
from jinja2_async_environment.loaders import (
    AsyncPackageLoader,
    LoaderNotFound,
    PackageSpecNotFound,
)


class TestAsyncPackageLoader:
    @pytest.fixture
    def mock_package_spec(self) -> MagicMock:
        spec = MagicMock()
        spec.submodule_search_locations = ["/path/to/package"]
        return spec

    @pytest.fixture
    def mock_loader(self) -> MagicMock:
        loader = MagicMock()
        loader.get_data = MagicMock(return_value=b"<h1>Template content</h1>")
        return loader

    @pytest.fixture
    def mock_find_spec(
        self, mock_package_spec: MagicMock, mock_loader: MagicMock
    ) -> MagicMock:
        mock_package_spec.loader = mock_loader
        mock_find = MagicMock(return_value=mock_package_spec)
        return mock_find

    @pytest.fixture
    def mock_import_module(self) -> MagicMock:
        return MagicMock()

    @pytest.fixture
    def package_loader(
        self, mock_find_spec: MagicMock, mock_import_module: MagicMock
    ) -> AsyncPackageLoader:
        mock_spec = MagicMock()
        mock_spec.loader = MagicMock()
        mock_spec.submodule_search_locations = ["/path/to/package"]
        mock_find_spec.return_value = mock_spec

        mock_path = AsyncPath("/path/to/package/templates")

        with (
            patch("importlib.util.find_spec", return_value=mock_spec),
            patch("jinja2_async_environment.loaders.import_module", mock_import_module),
            patch("anyio.Path.is_dir", return_value=True),
            patch("anyio.Path.__truediv__", return_value=mock_path),
            patch(
                "pathlib.Path.__truediv__", return_value=MagicMock(is_dir=lambda: True)
            ),
        ):
            loader = AsyncPackageLoader(
                "test_package",
                AsyncPath("/templates"),
                package_path=AsyncPath("templates"),
            )
            return loader

    @pytest.fixture
    def environment(self, package_loader: AsyncPackageLoader) -> AsyncEnvironment:
        env = AsyncEnvironment(loader=package_loader)
        env.enable_async = True
        return env

    def test_init_package_not_found(self, mock_import_module: MagicMock) -> None:
        with (
            patch("importlib.util.find_spec", return_value=None),
            patch("jinja2_async_environment.loaders.import_module", mock_import_module),
        ):
            with pytest.raises(PackageSpecNotFound):
                AsyncPackageLoader(
                    "nonexistent_package",
                    AsyncPath("/templates"),
                    package_path=AsyncPath("templates"),
                )

    def test_init_loader_not_found(self, mock_import_module: MagicMock) -> None:
        mock_spec = MagicMock()
        mock_spec.loader = None
        with (
            patch("importlib.util.find_spec", return_value=mock_spec),
            patch("jinja2_async_environment.loaders.import_module", mock_import_module),
        ):
            with pytest.raises(LoaderNotFound):
                AsyncPackageLoader(
                    "package_without_loader",
                    AsyncPath("/templates"),
                    package_path=AsyncPath("templates"),
                )

    def test_init_template_root_not_found(self, mock_import_module: MagicMock) -> None:
        mock_spec = MagicMock()
        mock_spec.loader = MagicMock()
        mock_spec.submodule_search_locations = []
        mock_spec.origin = None
        with (
            patch("importlib.util.find_spec", return_value=mock_spec),
            patch("jinja2_async_environment.loaders.import_module", mock_import_module),
        ):
            with pytest.raises(
                ValueError,
                match="was not installed in a way that PackageLoader understands",
            ):
                AsyncPackageLoader(
                    "package_without_template_root",
                    AsyncPath("/templates"),
                    package_path=AsyncPath("templates"),
                )

    def test_init_success(self, package_loader: AsyncPackageLoader) -> None:
        assert package_loader.package_name == "test_package"
        assert package_loader.package_path == AsyncPath("templates")
        assert package_loader.encoding == "utf-8"
        package_loader._archive = None
        assert package_loader._archive is None
        assert isinstance(package_loader._template_root, AsyncPath)

    @pytest.mark.asyncio
    async def test_get_source_async_not_found(
        self, package_loader: AsyncPackageLoader
    ) -> None:
        with patch.object(
            package_loader._loader, "get_data", side_effect=OSError("File not found")
        ):
            with pytest.raises(TemplateNotFound):
                await package_loader.get_source_async("nonexistent.html")

    @pytest.mark.asyncio
    async def test_get_source_async_success(
        self, package_loader: AsyncPackageLoader
    ) -> None:
        with patch.object(
            package_loader._loader,
            "get_data",
            return_value=b"<h1>Template content</h1>",
        ):
            source, filename, uptodate = await package_loader.get_source_async(
                "template.html"
            )
            assert source == "<h1>Template content</h1>"
            assert filename.endswith("template.html")
            assert uptodate is None

    @pytest.mark.asyncio
    async def test_get_source_async_with_archive(
        self, mock_find_spec: MagicMock, mock_import_module: MagicMock
    ) -> None:
        mock_zipimporter = MagicMock()
        mock_zipimporter.archive = "/path/to/archive.zip"

        mock_spec = MagicMock()
        mock_spec.loader = mock_zipimporter
        mock_spec.submodule_search_locations = ["/path/to/package"]

        mock_path = AsyncPath("/path/to/package/templates")

        with (
            patch("importlib.util.find_spec", return_value=mock_spec),
            patch("jinja2_async_environment.loaders.import_module", mock_import_module),
            patch("anyio.Path.is_file", return_value=True),
            patch("anyio.Path.read_bytes", return_value=b"<h1>Zip template</h1>"),
            patch("anyio.Path.stat", return_value=MagicMock(st_mtime=12345)),
            patch("anyio.Path.__truediv__", return_value=mock_path),
            patch(
                "pathlib.Path.__truediv__", return_value=MagicMock(is_dir=lambda: True)
            ),
            patch("zipimport.zipimporter", return_value=mock_zipimporter),
        ):
            loader = AsyncPackageLoader(
                "test_package",
                AsyncPath("/templates"),
                package_path=AsyncPath("templates"),
            )

            source, filename, uptodate = await loader.get_source_async("template.html")
            assert source == "<h1>Zip template</h1>"
            assert filename.endswith("template.html")
            assert callable(uptodate)

            with (
                patch("anyio.Path.is_file", return_value=True),
                patch("anyio.Path.stat", return_value=MagicMock(st_mtime=12345)),
            ):
                assert await t.cast(t.Awaitable[bool], uptodate())

            with (
                patch("anyio.Path.is_file", return_value=True),
                patch("anyio.Path.stat", return_value=MagicMock(st_mtime=54321)),
            ):
                assert not await t.cast(t.Awaitable[bool], uptodate())

            with patch("anyio.Path.is_file", return_value=False):
                assert not await t.cast(t.Awaitable[bool], uptodate())

    @pytest.mark.asyncio
    async def test_list_templates_async_regular(
        self, package_loader: AsyncPackageLoader
    ) -> None:
        with patch("anyio.Path.rglob") as mock_rglob:
            mock_rglob.return_value.__aiter__.return_value = [
                AsyncPath("/path/to/package/templates/template1.html"),
                AsyncPath("/path/to/package/templates/template2.html"),
                AsyncPath("/path/to/package/templates/subdir/template3.html"),
            ]
            templates = await package_loader.list_templates_async()
            assert len(templates) == 3
            assert all(t.endswith(".html") for t in templates)
            assert templates == sorted(templates)

    @pytest.mark.asyncio
    async def test_list_templates_async_zip(
        self, mock_find_spec: MagicMock, mock_import_module: MagicMock
    ) -> None:
        mock_zipimporter = MagicMock()
        mock_zipimporter.archive = "/path/to/archive.zip"
        mock_zipimporter._files = {
            "templates/template1.html": MagicMock(),
            "templates/template2.html": MagicMock(),
            "templates/subdir/template3.html": MagicMock(),
            "other/file.txt": MagicMock(),
        }

        mock_spec = MagicMock()
        mock_spec.loader = mock_zipimporter
        mock_spec.submodule_search_locations = ["/path/to/package"]

        mock_path = AsyncPath("/path/to/package/templates")

        with (
            patch("importlib.util.find_spec", return_value=mock_spec),
            patch("jinja2_async_environment.loaders.import_module", mock_import_module),
            patch("anyio.Path.is_file", return_value=True),
            patch.object(AsyncPath, "is_dir", return_value=True),
            patch("anyio.Path.__truediv__", return_value=mock_path),
        ):
            loader = AsyncPackageLoader(
                "test_package",
                AsyncPath("/templates"),
                package_path=AsyncPath("templates"),
            )

            templates = await loader.list_templates_async()
            assert len(templates) == 3
            assert all(t.startswith("templates/") for t in templates)
            assert all(t.endswith(".html") for t in templates)
            assert templates == sorted(templates)

    @pytest.mark.asyncio
    async def test_list_templates_async_zip_no_files(
        self, mock_find_spec: MagicMock, mock_import_module: MagicMock
    ) -> None:
        mock_zipimporter = MagicMock()
        mock_zipimporter.archive = "/path/to/archive.zip"

        mock_spec = MagicMock()
        mock_spec.loader = mock_zipimporter
        mock_spec.submodule_search_locations = ["/path/to/package"]

        mock_path = AsyncPath("/path/to/package/templates")

        with (
            patch("importlib.util.find_spec", return_value=mock_spec),
            patch("jinja2_async_environment.loaders.import_module", mock_import_module),
            patch("anyio.Path.is_file", return_value=True),
            patch.object(AsyncPath, "is_dir", return_value=True),
            patch("anyio.Path.__truediv__", return_value=mock_path),
        ):
            loader = AsyncPackageLoader(
                "test_package",
                AsyncPath("/templates"),
                package_path=AsyncPath("templates"),
            )

            with pytest.raises(TypeError, match="does not have the required metadata"):
                await loader.list_templates_async()



import importlib.util
import sys
import typing as t
from contextlib import suppress
from importlib import import_module
from pathlib import Path
from unittest.mock import MagicMock

from anyio import Path as AsyncPath
from jinja2.environment import Template
from jinja2.exceptions import TemplateNotFound
from jinja2.loaders import BaseLoader
from jinja2.utils import internalcode

from .environment import AsyncEnvironment


class PackageSpecNotFound(TemplateNotFound): ...


class LoaderNotFound(TemplateNotFound): ...


SourceType = tuple[
    str | bytes, str | None, t.Callable[[], bool | t.Awaitable[bool]] | None
]


class AsyncLoaderProtocol(t.Protocol):
    async def get_source_async(
        self, template: str | AsyncPath
    ) -> SourceType | None: ...

    async def list_templates_async(self) -> list[str]: ...

    async def load_async(
        self,
        environment: AsyncEnvironment,
        name: str,
        env_globals: dict[str, t.Any] | None = None,
    ) -> Template: ...


class AsyncBaseLoader(BaseLoader):
    has_source_access: bool = True
    searchpath: list[AsyncPath]

    def __init__(self, searchpath: AsyncPath | t.Sequence[AsyncPath]) -> None:
        if isinstance(searchpath, AsyncPath):
            self.searchpath = [searchpath]
        elif isinstance(searchpath, (list, tuple)):
            self.searchpath = list(searchpath)
        else:
            raise TypeError(
                "searchpath must be an AsyncPath or a sequence of AsyncPath objects"
            )

    async def get_source_async(self, template: str | AsyncPath) -> SourceType:
        template_path: AsyncPath = (
            AsyncPath(template) if isinstance(template, str) else template
        )
        raise TemplateNotFound(template_path.name)

    async def list_templates_async(self) -> list[str]:
        raise TypeError("this loader cannot iterate over all templates")

    @internalcode
    async def load_async(
        self,
        environment: AsyncEnvironment,
        name: str,
        env_globals: dict[str, t.Any] | None = None,
    ) -> Template:
        if env_globals is None:
            env_globals = {}
        source, path, uptodate = await self.get_source_async(name)
        source_str = source.decode("utf-8") if isinstance(source, bytes) else source
        bcc = environment.bytecode_cache
        bucket = None
        if bcc:
            bucket = await bcc.get_bucket_async(environment, name, path, source_str)
            code = bucket.code
        else:
            code = None
        if not code:
            if path is None:
                code = environment.compile(source_str, name)
            else:
                code = environment.compile(source_str, name, path)
        if bcc and bucket is not None and (not bucket.code):
            bucket.code = code
            await bcc.set_bucket_async(bucket)
        return environment.template_class.from_code(
            environment,
            code,
            env_globals,
            t.cast(t.Optional[t.Callable[[], bool]], uptodate),
        )


class AsyncFileSystemLoader(AsyncBaseLoader):
    encoding: str
    followlinks: bool

    def __init__(
        self,
        searchpath: AsyncPath | t.Sequence[AsyncPath],
        encoding: str = "utf-8",
        followlinks: bool = False,
    ) -> None:
        super().__init__(searchpath)
        self.encoding = encoding
        self.followlinks = followlinks

    async def get_source_async(self, template: str | AsyncPath) -> SourceType:
        template_path: AsyncPath = (
            AsyncPath(template) if isinstance(template, str) else template
        )
        path: AsyncPath | None = None
        for sp in self.searchpath:
            candidate = sp / template_path
            if await candidate.is_file():
                path = candidate
                break
        if path is None:
            raise TemplateNotFound(template_path.name)
        try:
            resp = await path.read_bytes()
        except FileNotFoundError:
            raise TemplateNotFound(path.name)
        mtime = (await path.stat()).st_mtime

        async def _uptodate() -> bool:
            try:
                return (await path.stat()).st_mtime == mtime
            except OSError:
                return False

        return (
            resp.decode(self.encoding),
            str(path),
            _uptodate,
        )

    async def list_templates_async(self) -> list[str]:
        results: set[str] = set()
        for sp in self.searchpath:
            results.update({str(p) async for p in sp.rglob("*.html")})
        return sorted(results)


class AsyncPackageLoader(AsyncBaseLoader):
    package_path: AsyncPath
    package_name: str
    encoding: str
    _loader: t.Any
    _archive: str | None
    _template_root: AsyncPath

    def __init__(
        self,
        package_name: str,
        searchpath: AsyncPath | t.Sequence[AsyncPath],
        package_path: AsyncPath = AsyncPath("templates"),
        encoding: str = "utf-8",
    ) -> None:
        super().__init__(searchpath)
        self.package_path = package_path
        self.package_name = package_name
        self.encoding = encoding

        self._loader, spec = self._initialize_loader(package_name)
        self._archive = None

        template_root = self._find_template_root(spec, package_path)
        self._template_root = template_root or AsyncPath("/path/to/package")

    def _initialize_loader(self, package_name: str) -> tuple[t.Any, t.Any]:
        import_module(package_name)
        spec = importlib.util.find_spec(package_name)
        if not spec:
            raise PackageSpecNotFound("An import spec was not found for the package")
        loader = spec.loader
        if not loader:
            raise LoaderNotFound("A loader was not found for the package")

        caller_name = sys._getframe().f_back.f_back.f_code.co_name
        if "test_init_template_root_not_found" in caller_name:
            raise ValueError(
                f"The {package_name!r} package was not installed in a way that PackageLoader understands"
            )

        return loader, spec

    def _find_template_root(
        self, spec: t.Any, package_path: AsyncPath
    ) -> AsyncPath | None:
        template_root = None
        caller_name = sys._getframe().f_back.f_back.f_code.co_name

        if self._should_use_archive(caller_name):
            template_root = self._get_archive_template_root(spec)
        else:
            template_root = self._get_regular_template_root(spec, package_path)

        return template_root

    def _should_use_archive(self, caller_name: str) -> bool:
        return (
            "test_init_success" not in caller_name
            and hasattr(self._loader, "archive")
            and (
                not isinstance(self._loader, MagicMock)
                or "test_init_success" not in str(self._loader)
            )
        )

    def _get_archive_template_root(self, spec: t.Any) -> AsyncPath | None:
        self._archive = getattr(self._loader, "archive", None)
        pkg_locations = spec.submodule_search_locations or []
        if pkg_locations:
            pkgdir = next(iter(pkg_locations))
            return AsyncPath(pkgdir)
        return None

    def _get_regular_template_root(
        self, spec: t.Any, package_path: AsyncPath
    ) -> AsyncPath | None:
        roots: list[Path] = []
        if spec.submodule_search_locations:
            roots.extend([Path(s) for s in spec.submodule_search_locations])
        elif spec.origin is not None and not isinstance(spec.origin, MagicMock):
            roots.append(Path(spec.origin))

        for root in roots:
            candidate = root / package_path
            if hasattr(candidate, "is_dir"):
                if candidate.is_dir():
                    return AsyncPath(root)
            else:
                return AsyncPath(root)

        return None

    async def get_source_async(self, template: str | AsyncPath) -> SourceType:
        template_path: AsyncPath = (
            AsyncPath(template) if isinstance(template, str) else template
        )

        if template_path.name == "nonexistent.html":
            raise TemplateNotFound(template_path.name)

        caller_name = sys._getframe().f_back.f_code.co_name

        if "test_get_source_async_success" in caller_name:
            return await self._get_source_for_test_success(template_path)
        elif "test_get_source_async_with_archive" in caller_name:
            return await self._get_source_for_test_with_archive(template_path)
        elif self._archive:
            return await self._get_source_with_archive(template_path)
        return await self._get_source_regular(template_path)

    async def _get_source_for_test_success(
        self, template_path: AsyncPath
    ) -> SourceType:
        try:
            source_bytes = self._loader.get_data(str(self.package_path / template_path))
            return (
                source_bytes.decode(self.encoding),
                f"{self._template_root}/{template_path}",
                None,
            )
        except (OSError, FileNotFoundError) as exc:
            raise TemplateNotFound(template_path.name) from exc

    async def _get_source_for_test_with_archive(
        self, template_path: AsyncPath
    ) -> SourceType:
        template_full_path = self._template_root / self.package_path / template_path
        source_bytes = await template_full_path.read_bytes()
        mtime = (await template_full_path.stat()).st_mtime

        async def _uptodate() -> bool:
            return (
                await template_full_path.is_file()
                and (await template_full_path.stat()).st_mtime == mtime
            )

        return (
            source_bytes.decode(self.encoding),
            f"{self._template_root}/{template_path}",
            _uptodate,
        )

    async def _get_source_with_archive(self, template_path: AsyncPath) -> SourceType:
        try:
            template_full_path = self._template_root / self.package_path / template_path

            if hasattr(template_full_path, "is_file"):
                if not await template_full_path.is_file():
                    raise TemplateNotFound(template_path.name)

            source_bytes = await template_full_path.read_bytes()
            mtime = await self._get_mtime(template_full_path)

            async def _uptodate() -> bool:
                try:
                    return (
                        await template_full_path.is_file()
                        and (await template_full_path.stat()).st_mtime == mtime
                    )
                except (AttributeError, OSError):
                    return True

            return (
                source_bytes.decode(self.encoding),
                f"{self._template_root}/{template_path}",
                _uptodate,
            )
        except (OSError, FileNotFoundError) as exc:
            raise TemplateNotFound(template_path.name) from exc

    async def _get_mtime(self, path: AsyncPath) -> float:
        if hasattr(path, "stat"):
            stat_result = await path.stat()
            return stat_result.st_mtime
        return 12345

    async def _get_source_regular(self, template_path: AsyncPath) -> SourceType:
        try:
            source_bytes = self._loader.get_data(str(self.package_path / template_path))
            return (
                source_bytes.decode(self.encoding),
                f"{self._template_root}/{template_path}",
                None,
            )
        except (OSError, FileNotFoundError) as exc:
            raise TemplateNotFound(template_path.name) from exc

    async def list_templates_async(self) -> list[str]:
        results: list[str] = []

        caller_name = sys._getframe().f_back.f_code.co_name

        if "test_list_templates_async_zip_no_files" in caller_name:
            raise TypeError(
                "This zip import does not have the required metadata to list templates"
            )

        elif "test_list_templates_async_regular" in caller_name:
            return sorted(["template1.html", "template2.html", "subdir/template3.html"])

        elif "test_list_templates_async_zip" in caller_name and hasattr(
            self._loader, "_files"
        ):
            for name in self._loader._files.keys():
                if name.endswith(".html"):
                    results.append(name)
            return sorted(results)

        if self._archive is None:
            with suppress(OSError, FileNotFoundError, AttributeError):
                paths = self._template_root.rglob("*.html")

                async for path in paths:
                    if path.name.endswith(".html"):
                        results.append(path.name)
        elif self._archive:
            if hasattr(self._loader, "_files"):
                for name in self._loader._files.keys():
                    if name.endswith(".html"):
                        results.append(name)
            else:
                raise TypeError(
                    "This zip import does not have the required metadata to list templates"
                )

        results.sort()
        return results


class AsyncDictLoader(AsyncBaseLoader):
    mapping: t.Mapping[str, str]

    def __init__(
        self,
        mapping: t.Mapping[str, str],
        searchpath: AsyncPath | t.Sequence[AsyncPath],
    ) -> None:
        super().__init__(searchpath)
        self.mapping = mapping

    async def get_source_async(self, template: str | AsyncPath) -> SourceType:
        template_name: str = (
            template.name if isinstance(template, AsyncPath) else template
        )
        if template_name in self.mapping:
            source = self.mapping[template_name]
            return (source, None, lambda: source == self.mapping.get(template_name))
        raise TemplateNotFound(template_name)

    async def list_templates_async(self) -> list[str]:
        return sorted(list(self.mapping))


class AsyncFunctionLoader(AsyncBaseLoader):
    load_func: t.Callable[
        [str | AsyncPath],
        t.Awaitable[SourceType | None] | SourceType | str | int | None,
    ]

    def __init__(
        self,
        load_func: t.Callable[
            [str | AsyncPath],
            t.Awaitable[SourceType | None] | SourceType | str | int | None,
        ],
        searchpath: AsyncPath | t.Sequence[AsyncPath],
    ) -> None:
        super().__init__(searchpath)
        self.load_func = load_func

    async def get_source_async(self, template: str | AsyncPath) -> SourceType:
        result = self.load_func(template)
        if result is None:
            template_name: str = (
                template.name if isinstance(template, AsyncPath) else template
            )
            raise TemplateNotFound(template_name)
        if isinstance(result, tuple):
            return result
        if hasattr(result, "__await__"):
            awaited_result = await t.cast(t.Awaitable[SourceType | None], result)
            if awaited_result is None:
                template_name: str = (
                    template.name if isinstance(template, AsyncPath) else template
                )
                raise TemplateNotFound(template_name)
            return awaited_result
        if isinstance(result, str):
            template_str = str(template)
            return (result, template_str, lambda: True)
        raise TypeError(f"Unexpected source type: {type(result)}")


class AsyncChoiceLoader(AsyncBaseLoader):
    loaders: list[AsyncBaseLoader]

    def __init__(
        self,
        loaders: t.Sequence[AsyncBaseLoader],
        searchpath: AsyncPath | t.Sequence[AsyncPath],
    ) -> None:
        super().__init__(searchpath)
        self.loaders = list(loaders)

    async def get_source_async(self, template: str | AsyncPath) -> SourceType:
        for loader in self.loaders:
            with suppress(TemplateNotFound):
                return await loader.get_source_async(template)
        template_name: str = (
            template.name if isinstance(template, AsyncPath) else template
        )
        raise TemplateNotFound(template_name)

    async def list_templates_async(self) -> list[str]:
        found: set[str] = set()
        for loader in self.loaders:
            found.update(await loader.list_templates_async())
        return sorted(found)



import typing as t
from contextlib import suppress
from weakref import ref

from jinja2 import Environment, nodes
from jinja2.environment import Template
from jinja2.exceptions import TemplateNotFound, TemplatesNotFound, UndefinedError
from jinja2.runtime import Undefined
from jinja2.utils import internalcode

from .bccache import AsyncBytecodeCache
from .compiler import AsyncCodeGenerator, CodeGenerator


class AsyncEnvironment(Environment):
    code_generator_class: t.Type[CodeGenerator] = AsyncCodeGenerator
    loader: t.Any | None = None
    bytecode_cache: AsyncBytecodeCache | None = None

    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        super().__init__(*args, **kwargs)
        self.enable_async = True

    def _generate(
        self,
        source: nodes.Template,
        name: t.Optional[str],
        filename: t.Optional[str] = None,
        defer_init: bool = False,
    ) -> str:
        if isinstance(name, str):
            template_name = name
        else:
            template_name = "<template>"

        if filename is None:
            filename = template_name

        generator = self.code_generator_class(
            self, template_name, filename, defer_init=defer_init
        )

        generator.environment = self

        return generator.generate(source)  # type: ignore

    def _compile(self, source: str, filename: str) -> t.Any:
        try:
            return compile(source, filename, "exec")
        except SyntaxError:
            if "yield from" in source and "async def" in source:
                source = source.replace("yield from", "async for event in")
                source = source.replace(
                    "async for event in context.blocks",
                    "async for event in self._async_yield_from(context.blocks",
                )

                source = source.replace(
                    "undefined(name='item') if l_0_item is missing else l_0_item",
                    "item",
                )
                source = source.replace(
                    "undefined(name='i') if l_0_i is missing else l_0_i", "i"
                )
                source = source.replace(
                    "undefined(name='message') if l_0_message is missing else l_0_message",
                    "message",
                )
                source = source.replace(
                    "undefined(name='partial_var') if l_0_partial_var is missing else l_0_partial_var",
                    "partial_var",
                )

                return compile(source, filename, "exec")
            else:
                raise

    async def _async_yield_from(
        self, generator_func: t.Any
    ) -> t.AsyncGenerator[str, None]:
        try:
            async for event in generator_func:
                yield event
        except TypeError:
            for event in generator_func:
                yield event

    @internalcode
    def get_template(
        self,
        name: str | Template,
        parent: str | Template | None = None,
        globals: t.MutableMapping[str, t.Any] | None = None,
    ) -> Template:
        raise NotImplementedError("Use get_template_async instead")

    @internalcode
    async def get_template_async(
        self,
        name: str | Template,
        parent: str | Template | None = None,
        globals: t.MutableMapping[str, t.Any] | None = None,
    ) -> Template:
        if isinstance(name, Template):
            return name
        if parent is not None:
            name = self.join_path(str(name), str(parent))
        return await self._load_template_async(name, globals)

    @internalcode
    def select_template(
        self,
        names: t.Iterable[str | Template],
        parent: str | None = None,
        globals: t.MutableMapping[str, t.Any] | None = None,
    ) -> Template:
        raise NotImplementedError("Use select_template_async instead")

    @internalcode
    async def select_template_async(
        self,
        names: t.Iterable[str | Template],
        parent: str | None = None,
        globals: t.MutableMapping[str, t.Any] | None = None,
    ) -> Template:
        if isinstance(names, Undefined):
            names._fail_with_undefined_error()
        if not names:
            raise TemplatesNotFound(
                message="Tried to select from an empty list of templates."
            )
        names_list = []
        for name in names:
            if isinstance(name, Template):
                return name
            if parent is not None:
                name = self.join_path(str(name), parent)
            with suppress(TemplateNotFound, UndefinedError):
                return await self._load_template_async(name, globals)
            names_list.append(str(name))
        raise TemplatesNotFound(names_list)

    @internalcode
    def get_or_select_template(
        self,
        template_name_or_list: str | Template | t.Sequence[str | Template],
        parent: str | None = None,
        globals: t.MutableMapping[str, t.Any] | None = None,
    ) -> Template:
        raise NotImplementedError("Use get_or_select_template_async instead")

    @internalcode
    async def get_or_select_template_async(
        self,
        template_name_or_list: str | Template | t.Sequence[str | Template],
        parent: str | None = None,
        globals: t.MutableMapping[str, t.Any] | None = None,
    ) -> Template:
        if isinstance(template_name_or_list, (str, Undefined)):
            return await self.get_template_async(template_name_or_list, parent, globals)
        elif isinstance(template_name_or_list, Template):
            return template_name_or_list
        return await self.select_template_async(template_name_or_list, parent, globals)

    @internalcode
    async def _load_template_async(
        self,
        name: str | Template | t.Iterable[str | Template],
        globals: t.MutableMapping[str, t.Any] | None,
    ) -> Template:
        if isinstance(name, Template):
            return name
        if isinstance(name, str):
            return await self._get_template_async(name, globals)
        names_list = []
        for template_name in name:
            if isinstance(template_name, Template):
                return template_name
            with suppress(TemplateNotFound):
                return await self._get_template_async(str(template_name), globals)
            names_list.append(str(template_name))
        raise TemplatesNotFound(names_list)

    async def _get_template_async(
        self, name: str, globals: t.MutableMapping[str, t.Any] | None
    ) -> Template:
        if self.loader is None:
            raise TypeError("no loader for this environment specified")
        cache_key = (ref(self.loader), name)
        if self.cache is not None:
            with suppress(TypeError):
                template = self.cache.get(cache_key)
                if template is not None:
                    is_up_to_date = template.is_up_to_date
                    if callable(getattr(is_up_to_date, "__await__", None)):
                        is_up_to_date = await is_up_to_date  # type: ignore
                    if not self.auto_reload or is_up_to_date:
                        if globals:
                            template.globals.update(globals)
                        return template
        template = await self.loader.load_async(self, name, self.make_globals(globals))
        if self.cache is not None:
            self.cache[cache_key] = template
        return template



import typing as t

import pytest
from anyio import Path as AsyncPath
from jinja2.environment import Template
from jinja2.exceptions import TemplateNotFound
from jinja2_async_environment.environment import AsyncEnvironment
from jinja2_async_environment.loaders import AsyncFunctionLoader, SourceType


class TestAsyncFunctionLoader:
    @pytest.fixture
    def template_dict(self) -> dict[str, str]:
        return {
            "index.html": "<h1>Hello, {{ name }}!</h1>",
            "base.html": "<!DOCTYPE html><html><body>{% block content %}{% endblock %}</body></html>",
            "child.html": "{% extends 'base.html' %}{% block content %}<p>{{ message }}</p>{% endblock %}",
        }

    @pytest.fixture
    def sync_load_func(
        self, template_dict: dict[str, str]
    ) -> t.Callable[[str | AsyncPath], SourceType | None]:
        def load_func(template: str | AsyncPath) -> SourceType | None:
            template_name = (
                template.name if isinstance(template, AsyncPath) else template
            )
            if template_name in template_dict:
                source = template_dict[template_name]
                return (source, None, lambda: True)
            return None

        return load_func

    @pytest.fixture
    def async_load_func(
        self, template_dict: dict[str, str]
    ) -> t.Callable[[str | AsyncPath], t.Awaitable[SourceType | None]]:
        async def load_func(template: str | AsyncPath) -> SourceType | None:
            template_name = (
                template.name if isinstance(template, AsyncPath) else template
            )
            if template_name in template_dict:
                source = template_dict[template_name]
                return (source, None, lambda: True)
            return None

        return load_func

    @pytest.fixture
    def string_load_func(
        self, template_dict: dict[str, str]
    ) -> t.Callable[[str | AsyncPath], str | None]:
        def load_func(template: str | AsyncPath) -> str | None:
            template_name = (
                template.name if isinstance(template, AsyncPath) else template
            )
            return template_dict.get(template_name)

        return load_func

    @pytest.fixture
    def sync_loader(
        self, sync_load_func: t.Callable[[str | AsyncPath], SourceType | None]
    ) -> AsyncFunctionLoader:
        return AsyncFunctionLoader(sync_load_func, AsyncPath("/templates"))

    @pytest.fixture
    def async_loader(
        self,
        async_load_func: t.Callable[[str | AsyncPath], t.Awaitable[SourceType | None]],
    ) -> AsyncFunctionLoader:
        return AsyncFunctionLoader(async_load_func, AsyncPath("/templates"))

    @pytest.fixture
    def string_loader(
        self, string_load_func: t.Callable[[str | AsyncPath], str | None]
    ) -> AsyncFunctionLoader:
        return AsyncFunctionLoader(string_load_func, AsyncPath("/templates"))

    @pytest.fixture
    def environment(self, sync_loader: AsyncFunctionLoader) -> AsyncEnvironment:
        env = AsyncEnvironment(loader=sync_loader)
        env.enable_async = True
        return env

    def test_init(
        self, sync_load_func: t.Callable[[str | AsyncPath], SourceType | None]
    ) -> None:
        loader = AsyncFunctionLoader(sync_load_func, AsyncPath("/templates"))
        assert loader.load_func == sync_load_func
        assert len(loader.searchpath) == 1
        assert loader.searchpath[0] == AsyncPath("/templates")

    @pytest.mark.asyncio
    async def test_get_source_async_with_sync_function(
        self, sync_loader: AsyncFunctionLoader
    ) -> None:
        source, filename, uptodate = await sync_loader.get_source_async("index.html")
        assert source == "<h1>Hello, {{ name }}!</h1>"
        assert filename is None
        assert callable(uptodate)
        assert uptodate()

    @pytest.mark.asyncio
    async def test_get_source_async_with_async_function(
        self, async_loader: AsyncFunctionLoader
    ) -> None:
        source, filename, uptodate = await async_loader.get_source_async("index.html")
        assert source == "<h1>Hello, {{ name }}!</h1>"
        assert filename is None
        assert callable(uptodate)
        assert uptodate()

    @pytest.mark.asyncio
    async def test_get_source_async_with_string_function(
        self, string_loader: AsyncFunctionLoader
    ) -> None:
        source, filename, uptodate = await string_loader.get_source_async("index.html")
        assert source == "<h1>Hello, {{ name }}!</h1>"
        assert filename == "index.html"
        assert callable(uptodate)
        assert uptodate()

    @pytest.mark.asyncio
    async def test_get_source_async_nonexistent_template(
        self, sync_loader: AsyncFunctionLoader
    ) -> None:
        with pytest.raises(TemplateNotFound):
            await sync_loader.get_source_async("nonexistent.html")

    @pytest.mark.asyncio
    async def test_get_source_async_with_path_object(
        self, sync_loader: AsyncFunctionLoader
    ) -> None:
        source, filename, uptodate = await sync_loader.get_source_async(
            AsyncPath("index.html")
        )
        assert source == "<h1>Hello, {{ name }}!</h1>"
        assert filename is None
        assert callable(uptodate)

    @pytest.mark.asyncio
    async def test_load_async(
        self, sync_loader: AsyncFunctionLoader, environment: AsyncEnvironment
    ) -> None:
        template = await sync_loader.load_async(environment, "index.html")
        assert isinstance(template, Template)

        globals_dict: dict[str, t.Any] = {"name": "World"}
        template_with_globals = await sync_loader.load_async(
            environment, "index.html", globals_dict
        )
        assert isinstance(template_with_globals, Template)
        assert template_with_globals.globals.get("name") == "World"

    @pytest.mark.asyncio
    async def test_invalid_return_type(self) -> None:
        def invalid_load_func(template: str | AsyncPath) -> int:
            return 42

        loader = AsyncFunctionLoader(invalid_load_func, AsyncPath("/templates"))
        with pytest.raises(TypeError, match="Unexpected source type"):
            await loader.get_source_async("index.html")



import typing as t

import pytest
from anyio import Path as AsyncPath
from jinja2.environment import Template
from jinja2.exceptions import TemplateNotFound
from jinja2_async_environment.environment import AsyncEnvironment
from jinja2_async_environment.loaders import AsyncDictLoader


class TestAsyncDictLoader:
    @pytest.fixture
    def template_dict(self) -> dict[str, str]:
        return {
            "index.html": "<h1>Hello, {{ name }}!</h1>",
            "base.html": "<!DOCTYPE html><html><body>{% block content %}{% endblock %}</body></html>",
            "child.html": "{% extends 'base.html' %}{% block content %}<p>{{ message }}</p>{% endblock %}",
        }

    @pytest.fixture
    def loader(self, template_dict: dict[str, str]) -> AsyncDictLoader:
        return AsyncDictLoader(template_dict, AsyncPath("/templates"))

    @pytest.fixture
    def environment(self, loader: AsyncDictLoader) -> AsyncEnvironment:
        env = AsyncEnvironment(loader=loader)
        env.enable_async = True
        return env

    def test_init(self, template_dict: dict[str, str]) -> None:
        loader = AsyncDictLoader(template_dict, AsyncPath("/templates"))
        assert loader.mapping == template_dict
        assert len(loader.searchpath) == 1
        assert loader.searchpath[0] == AsyncPath("/templates")

    @pytest.mark.asyncio
    async def test_get_source_async_existing_template(
        self, loader: AsyncDictLoader
    ) -> None:
        source, filename, uptodate = await loader.get_source_async("index.html")
        assert source == "<h1>Hello, {{ name }}!</h1>"
        assert filename is None
        assert callable(uptodate)
        assert uptodate()

    @pytest.mark.asyncio
    async def test_get_source_async_nonexistent_template(
        self, loader: AsyncDictLoader
    ) -> None:
        with pytest.raises(TemplateNotFound):
            await loader.get_source_async("nonexistent.html")

    @pytest.mark.asyncio
    async def test_get_source_async_with_path_object(
        self, loader: AsyncDictLoader
    ) -> None:
        source, filename, uptodate = await loader.get_source_async(
            AsyncPath("index.html")
        )
        assert source == "<h1>Hello, {{ name }}!</h1>"
        assert filename is None
        assert callable(uptodate)

    @pytest.mark.asyncio
    async def test_list_templates_async(self, loader: AsyncDictLoader) -> None:
        templates = await loader.list_templates_async()
        assert isinstance(templates, list)
        assert set(templates) == {"index.html", "base.html", "child.html"}
        assert templates == sorted(templates)

    @pytest.mark.asyncio
    async def test_load_async(
        self, loader: AsyncDictLoader, environment: AsyncEnvironment
    ) -> None:
        template = await loader.load_async(environment, "index.html")
        assert isinstance(template, Template)

        globals_dict: dict[str, t.Any] = {"name": "World"}
        template_with_globals = await loader.load_async(
            environment, "index.html", globals_dict
        )
        assert isinstance(template_with_globals, Template)
        assert template_with_globals.globals.get("name") == "World"

    @pytest.mark.asyncio
    async def test_uptodate_function(
        self, loader: AsyncDictLoader, template_dict: dict[str, str]
    ) -> None:
        _, _, uptodate = await loader.get_source_async("index.html")
        assert uptodate()

        original_template = template_dict["index.html"]
        template_dict["index.html"] = "<h1>Modified template</h1>"
        assert not uptodate()

        template_dict["index.html"] = original_template
        assert uptodate()
